/*
 * Copyright (C) 2005 Cesar Kallas
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * E-mail: cesarkallas@gmx.net
 * http://cesarkallas.soulivre.org
 *
 */
 
package estella.cliente;
import  estella.lib.*;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;

/**
* Funcoes do cliente
* @author Cesar Kallas
* @since 21/11/2005
*/
public class Cliente implements Runnable
{
	private Constantes CON;
	private Log LOG;
  private Tempo TEMPO;
  private StringTokenizer preServents;
	private Payload payload;
	private Descritor descritor;
	private Thread t_recebeDescritores;
	
	public Cliente()
	{
    LOG = new Log("Cliente");
    CON = new Constantes();
    TEMPO = new Tempo();
		preServents = new StringTokenizer(carregaServents(CON.LISTA_SERVENTS), ":");
	}
  
  public void run()
  {
		Tempo t = new Tempo();

		CON = new Constantes();
    LOG.fecha();

		Conexao conexao = conecta(preServents.nextToken(), Integer.parseInt(preServents.nextToken()));
		if(conexao != null) {
			synchronized(Dados.onlineServents) {
				Dados.onlineServents.adiciona(conexao);
			}
			t_recebeDescritores = new Thread(new recebeDescritores(conexao));
			t_recebeDescritores.start();
			//ping(conexao);
			query(conexao, 65123, "redes.txt");
		}
  }
  
	/**
	* Tenta conecta aos servents declarados na lista de servents pre-definidas em um arquivo
	* @return true ou false
	*/
  private Conexao conecta(String host, int porta)
  {
		try {
			// Tenta conectar, verifica resposta do servidor e se ok, adiciona a lista de conexoes
			Conexao conexao = new Conexao(host, porta);
			if(conexao.inicia()) {
				conexao.setTimeout(CON.SOCKET_TIMEOUT);
				conexao.envia(CON.GNUTELLA);
				if(conexao.recebe().indexOf(CON.GNUTELLA_OK) != -1) {
					conexao.setTimeout(0);
					LOG.info("Conectado ao servent " + conexao.status());
					return conexao;
				}
			}
			return null;
		}
		catch(Exception FalhaConectar) {
			LOG.falha("Nao consegui conectar no " + host + "@" + porta + " ->" + FalhaConectar.getMessage());
			return null;
		}
	}
	
	public boolean ping(Conexao conexao)
	{
		payload = new Payload();
		descritor = new Descritor(CON.PING, CON.TTL, 0, payload.ping());
		if(conexao.envia(new String(descritor.retorna())) == false)
			return false;
		return true;
	}
	
	public boolean query(Conexao conexao, int minVelocidade, String busca)
	{
		payload = new Payload();
		descritor = new Descritor(CON.QUERY, CON.TTL, 0, payload.query(minVelocidade, busca));
		if(conexao.envia(new String(descritor.retorna())) == false)
			return false;
		return true;
	}

	/**
	* Retorna uma lista de servents pre-definidas em um arquivo
	* @param caminho_servents String com o caminho para o arquivo com os servents. Arquivo definido no formato --  'servent:porta' --
	* @return String com todos os servents e portas encontrados no arquivo
	*/
	public String carregaServents(String caminho_servents)
	{
		BufferedReader lista_servents;
		String servent, servents;
			
		try {
			lista_servents = new BufferedReader(new FileReader(new File(caminho_servents)));
			servents = new String();
			servent = new String();

			while(servent != null) {
				servent = lista_servents.readLine();
				if(servent == null)
					break;
				servents += servent + ":";
			}

			LOG.info("Servents carregados..." + caminho_servents);
			lista_servents.close();
			return servents;
		}
		catch(Exception LerServents) {
			LOG.falha(LerServents.getMessage());
			return null;
		}
	}
}

/**
* Thread para tratar requisicao recebida pelo servidor
* @param sck_cliente Socket com a conexao do cliente
* @since 23/11/2005
* @author Cesar Kallas - cesarkallas at gmx . net
*/
class recebeDescritores implements Runnable
{
  private Conexao conexao;
  private Log LOG;
  private Constantes CON;
	private Descritor descritor;
	private String mensagem;
	private Dados dados;
	private Thread t_trataDescritor;
		
  public recebeDescritores(Conexao cliente)
  {
    LOG = new Log("Cliente");
    CON = new Constantes();
		conexao = cliente;
  }

  public void run()
  {
		LOG.info("Escutando descritores de " + conexao.status());
		// Tratar descritores que cheguem ao cliente
		while(conexao.status() != null) {
			try {
				mensagem = new String(conexao.recebe());
			}
			catch(Exception ReceberDescritor)
			{
				LOG.falha(ReceberDescritor.getMessage());
				return;
			}
			descritor = new Descritor(mensagem);
			t_trataDescritor = new Thread(new trataDescritor(descritor, conexao));
			t_trataDescritor.start();
		}
		synchronized(Dados.onlineServents) {
			Dados.onlineServents.remove(conexao);
		}
		LOG.info("Cliente desconectado");
		return;
	}
}

class trataDescritor implements Runnable
{
	private Descritor descritor;
	private Log LOG;
	private Constantes CON;
	private Conexao conexao;
	private Payload payload;
		
	public trataDescritor(Descritor descritor, Conexao conexao)
	{
		this.descritor = descritor;
		this.conexao = conexao;
		this.CON = new Constantes();
		this.LOG = new Log("Cliente");
	}
	
	public void run()
	{
		// Tamanho minimo de descritor 20 (id + ttl + hops + payload length)
		if(descritor.retorna().length < 20)
			return;
		
		switch(descritor.tipo())
		{
			case (byte)00000000: break;
			case (byte)00000001: { pong(); break; }
			case (byte)10000001: { queryhits(); break; }
			default: LOG.info("ho ho...o que eh isso?"); return;
		}
	}
	
	private void pong()
	{		
		LOG.info("pong recebido id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Tempo tempo = new Tempo();
		Payload payload = new Payload(CON.PONG, descritor.payload());
		Servent servent = new Servent(payload.pongIp(), payload.pongPorta(), payload.pongNumArqComp(), payload.pongTamArqComp(), tempo.data()+"  "+tempo.hora());
		
		synchronized(Dados.servents) {
			Dados.servents.adiciona(servent);
		}
		LOG.info("servent cadastrado IP: " + payload.pongIp() + " porta: " + payload.pongPorta() + " numArqComp: " + payload.pongNumArqComp() + " tamArqComp: " + payload.pongTamArqComp());
	}
	
	private void queryhits()
	{
		LOG.info("queryhits recebido id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Payload payload = new Payload(CON.QUERYHITS, descritor.payload());
		LOG.info("resultado da pesquisa - hits: " + payload.queryhitsNum() + " ip: " + payload.queryhitsIp() + " porta: " + payload.queryhitsPorta());

		byte[] resultset = payload.queryhitsResultset();
		String resultados = new String();
		String resultado;
		int off=0;
		
		System.out.println("tamanho do resultset: " + resultset.length);
		System.out.println("tamanho do queryhits: " + descritor.payload().length);
		
		payload.off=0;
		for(int i=0; i<payload.queryhitsNum(); i++) {
			resultado = new String(payload.retornaHit(resultset));
			resultados += resultado;
		}
	}
}
